(function() {

	function CustomizationService($rootScope, ScreenService, templateManager, RepositoryFactory, Query, ZhUtil, restEngine, ZHPromise, CustomizationsRequestsBuilder, ApplicationContext, containerMerger, eventAggregator, eventEngine, preferencesUserService, ZhLogger, ZhKeyboardService) {

		var self = this;
		var customizationsRepositories;
		var customizationsTypes = ["layout", "filter", "view"];
		var SUFFIX_ORIGINAL_VALUE = "original";
		var userPreference;

		function extendBehavior() {
			Util.extend(self, preferencesUserService);
		}
		extendBehavior();

		var validateRegister = function(register) {
			var valid = register && register.containerId && register.widgetId;
			if (!valid) {
				ZhLogger.warning('The container and the widget must have IDs to save customizations');
			}
			return valid;
		};

		var getCustomizationsRepositories = function() {
			return templateManager.projectPromise.then(function() {
				customizationsRepositories = templateManager.project.customizationsRepositories || [];
			});
		};

		var createCustomizationsRepositories = function() {
			customizationsRepositories.forEach(function(custoRepo) {
				var repositoryName = custoRepo.dbName + custoRepo.collectionName;
				ZhUtil.registerDataSource(repositoryName, "_id");
				repositoryName = "/" + repositoryName;
				custoRepo.repository = RepositoryFactory.factory(repositoryName, "LOCAL");
			});
		};

		self.fillCustomizationsRepositories = function() {
			var promises = customizationsRepositories.map(function(custoRepo) {
				var request = CustomizationsRequestsBuilder.buildRequest("GET", "findAll", custoRepo, []);
				return restEngine.request(request).then(function(responseData) {
					return custoRepo.repository.save(responseData.dataset.foundData);
				});
			});
			return ZHPromise.all(promises);
		};

		self.getRepository = function(custoId) {
			var customizationRepository = getCustomizationRepositoryById(custoId);
			if (!Util.isEmptyOrBlank(customizationRepository) && customizationRepository.repository) {
				return customizationRepository.repository;
			}
		};

		self.removeCustomizations = function(customizationsToDelete) {
			var customizationsMappedById = groupIdsByCustomizationRepository(customizationsToDelete);
			var promises = customizationsMappedById.map(function(customization) {
				var customizationRepository = getCustomizationRepositoryById(customization.custoRepo);
				var query = Query.build()
					.where("_id").in(customization.ids);
				var request = CustomizationsRequestsBuilder.buildRequest("DELETE", "remove", customizationRepository, createCustomizationsIdArrayIndexedById(customization.ids));
				return restEngine.request(request).then(function() {
					return customizationRepository.repository.remove(query);
				});
			});
			return ZHPromise.all(promises)
				.then(function(promisesResult) {
					eventAggregator.publish("OnUpdateCustomizationsRepositories");
					return promisesResult;
				});
		};

		self.saveCustomization = function(rows, customizationType) {
			rows = Util.parseToArray(rows);
			var promises = rows.map(function(row) {
				if (validateRegister(row) && customizationType && ~customizationsTypes.indexOf(customizationType)) {
					row.customizationType = row.customizationType || customizationType;
					var customizationRepository = getCustomizationRepositoryById(row.customizationRepository);
					var request = CustomizationsRequestsBuilder.buildRequest("POST", "save", customizationRepository, [row]);

					return restEngine.request(request).then(function(responseData) {
						var savedPreferences = responseData.dataset.response.saved;

						if (savedPreferences.length === 1) {
							row._id = savedPreferences[0]._id;
							delete row.unsync_id;
						}

						return customizationRepository.repository.save([row]);
					});
				} else {
					return ZHPromise.when({});
				}
			});

			return ZHPromise.all(promises)
				.then(function(promisesResult) {
					eventAggregator.publish("OnUpdateCustomizationsRepositories");
					return promisesResult;
				});
		};

		self.getWidgetCustomizations = function(widget, customizationType) {
			var promises = customizationsRepositories.map(function(repository) {
				var query = Query.build()
					.where("containerId").equals(widget.container.id)
					.where("widgetId").equals(widget.id)
					.where("customizationType").equals(customizationType);
				var userId = preferencesUserService.getPreferenceUserRegistered();
				if (userId && repository.customizationsByUser) {
					query.where("userId").equals(userId);
				}
				return repository.repository.find(query);
			});
			return ZHPromise.all(promises).then(function(reponses) {
				return reponses.reduce(function(accumulatedCustomizations, response) {
					return accumulatedCustomizations.concat(response);
				}, []);
			});
		};

		self.transferCustomizationsRepository = function(widget, rows, destinyRepository) {
			ScreenService.showLoader();
			return self.removeCustomizations(rows).then(function() {
				var syncIds = {};
				promises = rows.map(function(row) {
					return transferViewCustomizations(row, destinyRepository, widget).then(function() {
						row.customizationRepository = destinyRepository;
						row.unsync_id = new Date().getTime();
						removeNotInterestingProperties(row);
						delete row._id;
						return self.saveCustomization(row, row.customizationType).then(function(syncId) {
							syncIds[row.customizationType] = syncId;
						});
					});
				});
				return ZHPromise.all(promises).then(function() {
					ScreenService.hideLoader();
					return syncIds;
				});
			});
		};

		var transferViewCustomizations = function(view, destinyRepository, widget) {
			if (view.customizationType == "view") {
				var viewCustomizationsMapped = [{
					"customizationType": "filter",
					"id": view.filterId
				}, {
					"customizationType": "layout",
					"id": view.preferenceId
				}];
				var viewCustomizations = [];
				var customizationRepository = view.customizationRepository;
				var promises = viewCustomizationsMapped.map(function(customization) {
					return self.getWidgetCustomizationsByRepository(widget, customization.customizationType, customizationRepository, customization.id)
						.then(function(customizationObject) {
							viewCustomizations.push(customizationObject.shift());
						});
				});
				return ZHPromise.all(promises).then(function() {
					return self.transferCustomizationsRepository(widget, viewCustomizations, destinyRepository).then(function(syncIdsMapped) {
						view.filterId = syncIdsMapped.filter;
						view.preferenceId = syncIdsMapped.layout;
					});
				});
			} else {
				return ZHPromise.when();
			}
		};

		self.getCustomizationView = function(widget, customizationRepository, customizationType, id) {
			var idPropertyName = customizationType == "filter" ? "filterId" : "preferenceId";
			return self.getWidgetCustomizationsByRepository(widget, "view", customizationRepository)
				.then(function(viewObjects) {
					var viewObject;
					if (Util.isArray(viewObjects) && viewObjects.length > 0) {
						viewObject = viewObjects.filter(function(view) {
							return view[idPropertyName] == id;
						});
					}
					return viewObject;
				});
		};

		var removeNotInterestingProperties = function(object) {
			Object.keys(object).forEach(function(key) {
				if (~key.indexOf('__')) {
					delete object[key];
				}
			});
		};

		self.getWidgetCustomizationsByRepository = function(widget, customizationType, repositoryType, customizationId) {
			var customizationRepository = getCustomizationRepositoryById(repositoryType);
			var query = Query.build()
				.where("containerId").equals(widget.container.id)
				.where("widgetId").equals(widget.id)
				.where("customizationType").equals(customizationType);
			var userId = preferencesUserService.getPreferenceUserRegistered();
			if (userId && customizationRepository.customizationsByUser) {
				query.where("userId").equals(userId);
			}
			if (customizationId) {
				query.where("_id").equals(customizationId);
			}
			return customizationRepository.repository.find(query);
		};

		self.getCustomizationsRepositoriesTypes = function() {
			return customizationsRepositories;
		};

		self.getUpdatedCustomizationsRepositoriesTypes = function() {
			return self.fillCustomizationsRepositories().then(function() {
				return customizationsRepositories;
			});
		};

		self.getLabelRepositoryType = function(repositoryTypeValue) {
			return customizationsRepositories.filter(function(repository) {
				return repository.value == repositoryTypeValue;
			}).shift().label;
		};

		self.applyLayout = function(customization, editingWidget) {
			var eventData = createEventData(customization, editingWidget);
			if (self.layoutOnClick) {
				self.layoutOnClick(eventData);
			}
		};

		self.applyFilter = function(customization, editingWidget) {
			var eventData = createEventData(customization, editingWidget);
			if (self.filterOnClick) {
				self.filterOnClick(eventData);
			}
		};

		self.applyVisualization = function(customization, editingWidget) {
			var eventData = createEventData(customization, editingWidget);
			if (self.visualizationOnClick) {
				self.visualizationOnClick(eventData);
			}
		};

		function createEventData(customization, editingWidget) {
			return {
				"customization": customization,
				"editingWidget": editingWidget
			};
		}

		var groupIdsByCustomizationRepository = function(customizations) {
			var customizationsMappedById = [];
			customizationsRepositories.forEach(function(custoRepo) {
				var sameCustoRepoIds = customizations.filter(function(customization) {
					return customization.customizationRepository == custoRepo.value;
				});
				var map = {
					"custoRepo": custoRepo.value,
					"ids": sameCustoRepoIds.map(function(customization) {
						return customization._id;
					})
				};
				if (Util.isArray(map.ids) && map.ids.length > 0) {
					customizationsMappedById.push(map);
				}
			});
			return customizationsMappedById;
		};

		function createCustomizationsIdArrayIndexedById(customizationsId) {
			var customizationsToDelete = [];
			customizationsId.forEach(function(id) {
				customizationsToDelete.push({
					"_id": id
				});
			});
			return customizationsToDelete;
		}

		var getCustomizationRepositoryById = function(custoId) {
			return customizationsRepositories.filter(function(custoRepo) {
				return custoRepo.value == custoId;
			}).shift();
		};

		//-------------Begin Execution
		getCustomizationsRepositories().then(function() {
			createCustomizationsRepositories();
			self.fillCustomizationsRepositories().then(function() {
				eventAggregator.publish("OnUpdateCustomizationsRepositories");
			});
		});
	}

	Configuration(function(ContextRegister) {
		ContextRegister.register("zhCustomizationsService", CustomizationService);
	});

})();